import gym
from gym import logger, spaces
import numpy as np
from gym.envs.classic_control import rendering

class NavigationEnv(gym.Env):
	"""
	Description:
		N Agents cooperatively try to cover a set of N landmarks while avoiding colliding with other agents.
	Source:
		The environment appeared in the MADDPG paper (https://arxiv.org/pdf/1706.02275.pdf).
	Observation:
		Type: Box(n_agents, n_agents*4)
        Num    						 		    Observation               		  Min                             Max
        [i < n_agents, 2*j    <  n_agents*2]    Relative xPos agent to agent      -sqrt(2)*n_agents*world_size    sqrt(2)*n_agents*world_size
        [i < n_agents, 2*j+1  <  n_agents*2]    Relative yPos agent to agent      -sqrt(2)*n_agents*world_size    sqrt(2)*n_agents*world_size
        [i < n_agents, n_agents*2 <=    2*j]    Relative xPos agent to landmark   -sqrt(2)*n_agents*world_size    sqrt(2)*n_agents*world_size
        [i < n_agents, n_agents*2 <=  2*j+1]    Relative xPos agent to landmark   -sqrt(2)*n_agents*world_size    sqrt(2)*n_agents*world_size

	Actions:
		Type: Box(n_agents, 2)
        Num    						 		 	Observation               		  Min                             Max
		[i < n_agents, 0]						Angle of movement				  -pi							  pi
		[i < n_agents, 1]						Distance of movement			  0.0							  1.0

	Reward:
		 The agents are punished for uncovered landmarks and for colliding. The trade off between both can be adjusted with the parameter tau:
		 reward = - (tau * n_uncovered + (1-tau) * n_colliding)
 
				
	Starting State:
		 Agents start at random position in (2.5*N)x(2.5*N) units big environment one agents diameter is 1 unit.

	Episode Termination:
		 Either after max_steps or after keeping all landmarks covered for hold_steps.
	"""

	metadata = {"render.modes": ["human"]}

	def __init__(self, n_agents, world_size: float = 2.5, tau: float = 0.5, max_steps: int = np.inf,
                 hold_steps: int = 10, sparse: bool = False):
		
		self.border = world_size*n_agents
		self.n_agents = n_agents
		self.agents = np.random.rand(n_agents, 2)*(self.border - 1) + 0.5
		self.landmarks = np.random.rand(n_agents, 2)*(self.border - 0.5) + 0.25
		self.tau = tau
		self.sparse = sparse
		
		self.low_state = np.array(
        	[[-self.border for j in range(n_agents*4 - 2)] for i in range(n_agents)], dtype=np.float32
        )

		self.high_state = np.array(
        	[[self.border for j in range(n_agents*4 - 2)] for i in range(n_agents)], dtype=np.float32
        )

		self.action_space = spaces.Box(
			low=np.array([[-np.pi, 0.0]]*n_agents), high=np.array([[np.pi, 1.0]]*n_agents), dtype=np.float32
        )
		
		self.observation_space = spaces.Box(
			low=self.low_state, high=self.high_state, dtype=np.float32
        )

		self.done = False
		self.step_counter = 0
		self.max_steps = max_steps
		
		self.hold_step_counter = 0
		self.hold_steps = hold_steps

		self.viewer = None


	def step(self, act: np.array):
		
		if self.done:
			logger.warn(
				"You are calling 'step()' even though this "
				"environment has already returned done = True. You "
				"should always call 'reset()' once you receive 'done = "
				"True' -- any further steps are undefined behavior."
			)

		self._check_action(act)
		trans = [[r*np.cos(phi), r*np.sin(phi)] for phi, r in act]
		self.agents += trans
		self.agents = np.clip(self.agents, 0.5, self.border-0.5)
		obs, dist = self._get_observations()
		
		# number of agent collistions (all distances between two agents smaller than agent diameter. Divided by two due to every distance twice in matrix) 
		n_collisions = (np.count_nonzero(dist[:, :self.n_agents - 1] < 1))//2
		
		# number of uncovered landmarks
		n_uncovered = np.count_nonzero(np.all(dist[:, self.n_agents - 1:] > 0.25, axis=0))
		
		# sum of distances of closest agent for each landmark
		min_dist = np.sum(np.min(dist[:, self.n_agents - 1:], axis=0))
		
		if self.sparse:
			rew = - (self.tau * n_uncovered + (1-self.tau) * n_collisions)
	
		else:
			rew = - (self.tau * min_dist + (1-self.tau) * n_collisions)
		

		self.step_counter += 1

		if n_uncovered == 0:
			self.hold_step_counter += 1
		
		else:
			self.hold_step_counter = 0

		if self.step_counter >= self.max_steps or self.hold_step_counter >= self.hold_steps:
			self.done = True
		
		return obs, rew, self.done, {}


	def reset(self, reset_positions: bool = True) -> np.array:
		
		if reset_positions:	
			self.agents = np.random.rand(self.n_agents, 2)*(self.border - 1) + 0.5
			self.landmarks = np.random.rand(self.n_agents, 2)*(self.border - 0.5) + 0.25

		self.done = False
		self.step_counter = 0
		self.hold_step_counter = 0
		obs, _ = self._get_observations()
		
		return obs


	def _check_action(self, action):
		
		if action.shape != (self.n_agents, 2):
			raise ValueError(f"Action has wrong shape. Expected ({self.n_agents}, 2), got {action.shape}.")
				

	def _get_observations(self) -> np.array:
		# n_agents x (n_agents - 1 + n_landmarks)*2
		observations = np.zeros((self.n_agents, (2*self.n_agents - 1)*2))
		# n_agents x (n_agents - 1 + n_landmarks)
		dist = np.zeros((self.n_agents, 2*self.n_agents - 1))

		for i in range(self.n_agents):
			for j in range(self.n_agents):
				
				# agent-landmark x, y distances
				observations[i, (self.n_agents-1)*2 + 2*j:(self.n_agents-1)*2 + 2*j + 2] = self.agents[i] - self.landmarks[j]
				
				# agent-landmark direct distances
				dist[i, self.n_agents-1 + j] = np.linalg.norm(self.agents[i] - self.landmarks[j], ord=1)
				
				# agent-agent not with itself
				if i == j:
					continue
				
				k = j if j < i else j - 1

				# agent-agent x, y distances
				observations[i, 2*k:2*k+2] = self.agents[i] - self.agents[j]
				
				# agent-landmark direct distances
				dist[i, k] = np.linalg.norm(self.agents[i] - self.agents[j], ord=1)
		
		return observations, dist

	# Render environment
	def render(self, mode="human"):
		
		if self.viewer == None:
			length = int(self._world_to_render(self.border))
			self.viewer = rendering.Viewer(width=length, height=length)
			self.viewer.set_bounds(0, self.border, 0, self.border)
			for i in range(len(self.landmarks)):
				landmark = rendering.make_circle(radius=0.25)
				landmark.set_color(1, 0, 0)
				landmark.add_attr(rendering.Transform(translation=(self.landmarks[i][0], self.landmarks[i][1])))
				self.viewer.add_geom(landmark)

		for i in range(len(self.agents)):
			agent = rendering.make_circle(radius=0.5)
			agent.set_color(0, 1, 0)	
			agent.add_attr(rendering.Transform(translation=(self.agents[i][0], self.agents[i][1])))
			self.viewer.add_onetime(agent)
		
		self.viewer.render(return_rgb_array=mode == "rgb_array")

	
	def _world_to_render(self, coordinates):
		if isinstance(coordinates, int) or isinstance(coordinates, float):
			return 100*coordinates//4
		elif isinstance(coordinates, list):
			return [c*100//4 for c in coordinates]
		elif isinstance(coordinates, tuple):
			return (c*100//4 for c in coordinates)

	def close(self):
		if self.viewer:
			self.viewer.close()
			self.viewer = None

	def get_obs_dim(self):
		return self.n_agents*4 - 2

	def get_act_dim(self):
		return 2

if __name__ == "__main__":
	n_agents = 3
	env = NavigationEnv(n_agents)
	obs = env.reset()
	env.render()
	while True:
		obs, r, done, _ = env.step(env.action_space.sample())
		#env.step(np.array([[6.28, 0.2]]*n_agents), normalize=True)
		env.render()



import gym
from gym import logger
import numpy as np

class TwoStepEnv(gym.Env):
	"""
	Description:
		Two Agents play a tabular game with a shared reward. In the first round the first agent can decide to
		either play one of the tabular games 2A or 2B in the next round.
	Source:
		The environment appeared in the QMIX paper.
	Observation:
		Type: MultiDiscrete(2)
		Num    Observation					0			  1				2
		0	   Agent 1 label			  Agent 1		Agent 2			?
		1	   Agent 2 label			  Agent 1		Agent 2			?
	Actions:
		Type: Discrete(2)
		Num    Action
		0	   Action 0
		1	   Action 1
	Reward:
		  
		  (0/0) | (0/1)    (1/0) | (1/1)
				/				 \
		  2A 0 1				2B 0 1
		   0|7|7|				 0|0|1|
		   1|7|7|				 1|1|8|

				
	Starting State:
		 Both agents recieve their respective label in the beginning in the second round only 2 for unknown.

	Episode Termination:
		 After two rounds
	"""

	metadata = {"render.modes": ["human"]}

	def __init__(self):
		self.game_state = 0

	def step(self, act):
		
		# unknown observation
		obs = [[2], [2]]

		if self.game_state == 0:
			# next game state
			if act[0] == 0:
				self.game_state = 1
			elif act[0] == 1:
				self.game_state = 2
			
			done = 0
			rew = 0
		
		elif self.game_state == 1:
			done = 1
			rew = 7
			
			# next game state
			self.game_state = 3 + 2*int(act[0]) + int(act[1])

		elif self.game_state == 2:
			done = 1
			if act[0] == 0 and act[1] == 0:
				rew = 0
			elif act[0] == 1 and act[1] == 1:
				rew = 8
			elif act[0] == 0 and act[1] == 1:
				rew = 1
			elif act[0] == 1 and act[1] == 0:
				rew = 1
			
			# next game state
			self.game_state = 7 + 2*int(act[0]) + int(act[1])
		
		# game is over yet step is called
		elif self.game_state >= 3:
			done = 1
			rew = 0
			logger.warn(
				"You are calling 'step()' even though this "
				"environment has already returned done = True. You "
				"should always call 'reset()' once you receive 'done = "
				"True' -- any further steps are undefined behavior."
			)

		return np.array(obs), rew, done, {}

	def reset(self):
		self.game_state = 0
		obs = [[0], [1]]
		return np.array(obs)

	def render(self, mode="human"):
		l1 = "7" if self.game_state == 0 or self.game_state == 1 or self.game_state == 3 else "█"
		l2 = "7" if self.game_state == 0 or self.game_state == 1 or self.game_state == 4 else "█"
		l3 = "7" if self.game_state == 0 or self.game_state == 1 or self.game_state == 5 else "█"
		l4 = "7" if self.game_state == 0 or self.game_state == 1 or self.game_state == 6 else "█"
		r1 = "0" if self.game_state == 0 or self.game_state == 2 or self.game_state == 7 else "█"
		r2 = "1" if self.game_state == 0 or self.game_state == 2 or self.game_state == 8 else "█"
		r3 = "1" if self.game_state == 0 or self.game_state == 2 or self.game_state == 9 else "█"
		r4 = "8" if self.game_state == 0 or self.game_state == 2 or self.game_state == 10 else "█"
		print(f"|{l1}|{l2}|   |{r1}|{r2}|\n|{l3}|{l4}|	 |{r3}|{r4}|\n")



import gym

class NormalizeObsWrapper(gym.ObservationWrapper):

    def __init__(self, env):
        super().__init__(env)
        self.scale = 1.0/(env.observation_space.high - env.observation_space.low)
        self.bias = - env.observation_space.low

    def observation(self, obs):
        obs = (obs + self.bias)*self.scale
        return obs

class NormalizeActWrapper(gym.ActionWrapper):

    def __init__(self, env):
        super().__init__(env)
        self.scale = (env.action_space.high - env.action_space.low)
        self.bias =  self.action_space.low

    def action(self, act):
        act = act*self.scale + self.bias
        return act

class NormalizeRewWrapper(gym.RewardWrapper):

    def __init__(self, env):
        super().__init__(env)
        self.scale = 1.0/(2*env.n_agents)

    def reward(self, rew):
        # modify rew
        return rew*self.scale



from setuptools import setup, find_packages
from marlenvs import __version__

setup(
		name='marlenvs',
		version=__version__,
		author='Lukas König',
		author_email='lukasmkoenig@gmx.net',
        packages=find_packages(),
		python_requires='>=3.9',
		install_requires=['gym', 'numpy', 'wheel', 'pyglet'],
		decription="Multi-Agent Reinforcement Learning environments for gym."
	)
